/*-------------------------------------------------------------------------
 *
 * ctpp.hpp
 *
 *         : BSD.
 *         Copyright (c) 2004 Andrei V. Shetuhin
 *
 *           CPP Template.
 *
 *
 *-------------------------------------------------------------------------
 */
#ifndef _CTPP_HPP__
#define _CTPP_HPP__ 1

// C Includes

// C++ Includes
#include <vector>
#include <string>
#include <map>
#include <stdexcept>

// Local Includes
#include <util/d2str.hpp>
#include <ctpp/param_data.hpp>
#include <ctpp/udf_fn.hpp>
#include <functions/std_fn_list.hpp>

namespace template_parser_ns
{
//  
enum e_token_type{TMPL_TEXT, TMPL_LOOP, TMPL_IF, TMPL_ELSE, TMPL_UNLESS, TMPL_VAR};

class t_template;

//
//  
//
struct function_param_data
{
	std::string             param;
	bool                    is_variable;
};

//
// 
//
struct stack_ref
{
	//  
	std::string                      elem_name;
	//   
	t_template                     * template_elem;
	// 
	udf_fn                         * function;
	//    .
	std::vector<function_param_data> function_parameters;
};
typedef std::vector<stack_ref> v_stack;

//
//  
//
struct template_ret_type
{
	//   
	e_token_type          token_type;
	//  
	std::string::iterator parse_pos;
	//  
	int                   line;
	//  
	int                   column;
};

//
//   
//
class t_template
{
public:
	//  
	virtual e_token_type get_type() = 0;

	//  
	virtual std::string & output() = 0;

	//   
	virtual v_stack & child_list() = 0;

	//  
	virtual void param(param_data * pParamData, param_data * pIRootParamData) = 0;

	//   
	virtual template_ret_type parse(std::string::iterator itmData, std::string::iterator itmDataEnd);

	//   
	virtual void additional_param(param_data * pParamData);

	// 
	virtual ~t_template() { ;; }
};

//
//  
//
class template_text:
  public t_template
{
public:
	// 
	template_text(template_parser_ns::udf_fn_factory * pIFactory = NULL, const int iILine = 1, const int iIPos = 1, const int iITabLength = 9, const bool bIDebug = false, const bool bIStrict = true, const bool bILoopContextVars = false, const bool bIGlobalVars = false);

	//  Factory
	udf_fn_factory * get_fn_factory();

	//  
	e_token_type get_type();

	//  
	std::string & output();

	//   
	v_stack & child_list();

	//  
	void param(param_data * pIParamData, param_data * pIRootParamData);

	//   
	template_ret_type parse(std::string::iterator itmData, std::string::iterator itmDataEnd);

	// 
	~template_text();
private:
	//  
	v_stack          vStack;
	//  
	std::string      sTextData;
	// 
	param_data     * pParamData;
	//  
	param_data     * pRootParamData;

	// 
	udf_fn_factory * pFactory;
	//  
	int              iLine;
	//  
	int              iPos;
	//  
	int              iTabLength;
	// 
	bool             bDebug;
	//    
	bool             bStrict;
	//  
	bool             bLoopContextVars;
	//  
	bool             bGlobalVars;
	// Factory    
	bool             bSelfCreatedFactory;

	// ,     
	void             do_rollback_token(std::string::iterator itmData, std::string::iterator itmRollBackPos);
	//   
	void             fatal_parsing_error(const e_token_type &eExpectedToken, const e_token_type &eFoundToken);
	// enum -> Human Readable
	std::string      token2str(const e_token_type &eToken);
};


//
// 
//
class template_var:
  public t_template
{
public:
	// 
	template_var(udf_fn * pIFunction = NULL);

	//  
	e_token_type get_type();

	//  
	std::string & output();

	//   
	v_stack & child_list();

	//  
	void param(param_data * pParamData, param_data * pIRootParamData);

	// 
	~template_var();
private:
	std::string   sValue;
	std::string   sTmpValue;
	//  
	udf_fn      * pFunction;
};

//
// 
//
class template_loop:
  public t_template
{
public:
	// 
	template_loop(template_parser_ns::udf_fn_factory * pIFactory, const int iILine, const int iIPos, const int iITabLength, const bool bIDebug, const bool bIStrict, const bool bILoopContextVars, const bool bIGlobalVars);

	//  
	e_token_type get_type();

	//  
	std::string & output();

	//   
	v_stack & child_list();

	//  
	void param(param_data * pIParamData, param_data * pIRootParamData);

	//   
	template_ret_type parse(std::string::iterator itmData, std::string::iterator itmDataEnd);

	// 
	~template_loop();
private:
	//  
	t_template     * pTemplateText;

	// 
	param_data     * pParamData;
	//  
	param_data     * pRootParamData;

	//  
	std::string     sTextData;

	// 
	udf_fn_factory * pFactory;
	//  
	int              iLine;
	//  
	int              iPos;
	//  
	int              iTabLength;
	// 
	bool             bDebug;
	//    
	bool             bStrict;
	//  
	bool             bLoopContextVars;
	//  
	bool             bGlobalVars;
};

//
//  
//
class template_if:
  public t_template
{
public:
	// 
	template_if(template_parser_ns::udf_fn_factory * pIFactory, const int iILine, const int iIPos, const int iITabLength, const bool bIDebug, const bool bIStrict, const bool bILoopContextVars, const bool bIGlobalVars, const bool bIIsIf);

	//  
	e_token_type get_type();

	//  
	std::string & output();

	//   
	v_stack & child_list();

	//  
	void param(param_data * pIParamData, param_data * pIRootParamData);

	//   
	void additional_param(param_data * pIIfParamData);

	//   
	template_ret_type parse(std::string::iterator itmData, std::string::iterator itmDataEnd);

	// 
	~template_if();

private:
	//    
	t_template     * pTemplateIfText;
	//    
	t_template     * pTemplateElseText;

	// 
	param_data     * pParamData;
	//  
	param_data     * pRootParamData;
	//  
	param_data     * pIfParamData;

	//  
	std::string      sTextData;

	// 
	udf_fn_factory * pFactory;
	//  
	int              iLine;
	//  
	int              iPos;
	//  
	int              iTabLength;
	// 
	bool             bDebug;
	//    
	bool             bStrict;
	//  
	bool             bLoopContextVars;
	//  
	bool             bGlobalVars;
	//   IF
	bool             bIsIf;
};

} // namespace template_parser_ns
#endif // _CTPP_HPP__
// End.
